
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/time.h>
#include "input.h"
#include "pnginput.h"
#include "xoutput.h"
#include "tools.h"
#include "leetft.h"
#include "speedy.h"
#include "rtctimer.h"
#include "vidmode.h"

/**
 * Set this to 1 to see lots of debug messages.
 */
static int verbose = 0;

/**
 * Decrease this number to get better smoothness at the cost of
 * performance.  16 seems to be a good number here for smoothness.
 * Don't use a number less than 2.
 */
static int resync_frame = 256;

/**
 * Pixels to move per second.
 */
static int pixels_per_second = 400;

/**
 * Set this to 1 if the flat panel you're using isn't giving reasonable
 * modeline information.
 */
static int believe_measured_rate = 0;


static int timediff( struct timeval *large, struct timeval *small )
{
    return (   ( ( large->tv_sec * 1000 * 1000 ) + large->tv_usec )
             - ( ( small->tv_sec * 1000 * 1000 ) + small->tv_usec ) );
}

static double intensity_to_voltage_srgb( double val )
{
    /* Handle invalid values before doing a gamma transform. */
    if( val < 0.0 ) return 0.0;
    if( val > 1.0 ) return 1.0;

    /* sRGB uses an exponent of 2.4 in its nonstandard curve, but it
     *      * still advertises a display gamma of 2.2.
     *           */
    if( val <= 0.0031308 ) {
        return 12.92 * val;
    }
    return ( ( 1.055 * pow( val, 1.0 / 2.4 ) ) - 0.055 );
}

static int crawlpage = 0;
static int crawlpos = 0;
static pnginput_t *curpng = 0;

static int load_next_png( void )
{
    char filename[ 1024 ];

    if( curpng ) pnginput_delete( curpng );
    sprintf( filename, "/tmp/crawl%08d.png", crawlpage++ );
    curpng = pnginput_new( filename );
    if( !curpng ) {
        crawlpage = 0;
        sprintf( filename, "/tmp/crawl%08d.png", crawlpage++ );
        curpng = pnginput_new( filename );
        if( !curpng ) {
            fprintf( stderr, "viewcrawl: Can't find %s: nothing to draw.\n", filename );
            return 0;
        }
    }
    crawlpos = 0;

    return 1;
}

static int fill_next_page( unsigned char *data, int width, int height, int stride )
{
    int curpos = 0;
    int startpos = 0;
    int i;
    int done = 0;

    while( curpos < width ) {
        int curwidth = pnginput_get_width( curpng ) - crawlpos;
        int loadnext = 1;

        startpos = crawlpos;
        if( curwidth >= ( width - curpos ) ) {
            curwidth = ( width - curpos );
            crawlpos += curwidth;
            loadnext = 0;
            done = 1;
        }

        for( i = 0; i < height; i++ ) {
            unsigned char *curscanline = data + (i * stride) + (curpos * 4);

            if( i >= pnginput_get_height( curpng ) ) {
                blit_colour_packed4444_scanline( curscanline, curwidth, 0, 0, 0, 0 );
            } else {
                rgba32_to_argb32_scanline( curscanline,
                                           pnginput_get_scanline( curpng, i ) + (startpos * 4),
                                           curwidth );
            }
        }
        curpos += curwidth;

        if( loadnext ) {
            if( !load_next_png() ) {
                return 0;
            }
        }
    }
    return 1;
}

int main( int argc, char **argv )
{
    output_api_t *output = xoutput_get_output();
    input_t *in = input_new();
    int width, height, stride;
    unsigned char *page1;
    unsigned char *page2;
    struct timeval starttime;
    struct timeval lastblit;
    struct timeval curtime;
    vidmode_t *vidmode;
    int pos = 0;
    int pagewidth, pageheight, pagestride;
    int usecperframe;
    int refreshtime;
    int blittime = 0;
    int errortime = 0;
    int accel_step;
    int accel_max;
    int accel = 0;
    int paused = 1;
    rtctimer_t *rtctimer;
    int framecount = 0;
    int i;
    char c;

    while( (c = getopt( argc, argv, "vbs:p:" )) != -1 ) {
        switch( c ) {
        case 's': resync_frame = atoi( optarg ); break;
        case 'v': verbose = 1; break;
        case 'p': pixels_per_second = atoi( optarg ); break;
        case 'b': believe_measured_rate = 1; break;
        default:
            return 0;
        }
    }

    setup_speedy_calls( 1 );

    if( !load_next_png() ) {
        return 1;
    }

    if( !set_realtime_priority( 0 ) ) {
        fprintf( stderr, "viewcrawl: Can't set realtime priority, must be root.\n" );
        fprintf( stderr, "viewcrawl: Accuracy will suffer.\n" );
    }
    rtctimer = rtctimer_new( 1 );
    if( !rtctimer ) {
        fprintf( stderr, "viewcrawl: Can't open /dev/rtc needed for timing.\n" );
        fprintf( stderr, "viewcrawl: Defaulting to usleep accuracy (10ms).\n" );
    } else {
        if( !rtctimer_set_interval( rtctimer, 1024 ) ) {
            fprintf( stderr, "viewcrawl: Can't get millisecond accuracy from /dev/rtc.\n" );
            fprintf( stderr, "viewcrawl: You must be root or change max-user-freq in /proc.\n" );
            rtctimer_delete( rtctimer );
            rtctimer = 0;
        } else {
            rtctimer_start_clock( rtctimer );
            errortime = rtctimer_get_usecs( rtctimer );
        }
    }

    width = 800;
    height = 40;
    output->init( width, height );
    stride = output->get_output_stride();

    pagewidth = width;
    pageheight = height;
    pagestride = pagewidth * 4;

    page1 = (unsigned char *) malloc( pagestride * height );
    page2 = (unsigned char *) malloc( pagestride * height );
    if( !page1 || !page2 ) {
        fprintf( stderr, "viewcrawl: Can't allocte memory.\n" );
        return 1;
    }

    if( !fill_next_page( page1, pagewidth, pageheight, pagestride ) ) {
        fprintf( stderr, "viewcrawl: Can't open next crawl file, exiting.\n" );
    }
    if( !fill_next_page( page2, pagewidth, pageheight, pagestride ) ) {
        fprintf( stderr, "viewcrawl: Can't open next crawl file, exiting.\n" );
    }

    output->lock_output_buffer();
    memset( output->get_output_buffer(), 0, stride * height );
    output->unlock_output_buffer();
    output->show_frame();

    /* Try to measure refresh rate. */
    vidmode = vidmode_new();
    if( vidmode ) {
        double refresh = vidmode_get_refresh( vidmode );
        fprintf( stderr, "viewcrawl: Report refresh rate is %.2fhz.\n", refresh );
        refreshtime = (int) (((1.0 / refresh) * 1000.0 * 1000.0) + 0.5);
    } else {
        refreshtime = (int) (((1.0 / 85.0) * 1000.0 * 1000.0) + 0.5);
    }

    /* Blit a frame every refresh time. */
    usecperframe = refreshtime;

    /* Start with 200 pixels per second. */
    fprintf( stderr, "viewcrawl: Speed is %d pixels per second.\n", pixels_per_second );
    accel_max = ((((pixels_per_second * usecperframe) / 1000))*65536)/1000;
    accel_step = accel_max / 8;

    /* Reset our start time. */
    fprintf( stderr, "viewcrawl: VidMode says %d usec per frame.\n", usecperframe );

    {
        struct timeval bef, af;
        int sum = 0;
        int trials = 20;
        output->wait_for_retrace();
        output->wait_for_retrace();
        for( i = 0; i < trials; i++ ) {
            gettimeofday( &bef, 0 );
            output->wait_for_retrace();
            gettimeofday( &af, 0 );
            sum += timediff( &af, &bef );
        }
        fprintf( stderr, "viewcrawl: Roughly measured at %d usec over %d trials.\n",
                 sum / trials, trials );
        if( believe_measured_rate ) {
            fprintf( stderr, "viewcrawl: Assuming measured value is correct.\n" );
            usecperframe = sum / trials;
        }
    }

    output->wait_for_retrace();
    gettimeofday( &starttime, 0 );
    lastblit = starttime;
    framecount = 0;

    for(;;) {
        struct timeval beforeblit;
        struct timeval afterblit;

        output->poll_events( in );
        if( input_quit_requested( in ) ) break;
        if( input_pause_requested( in ) ) paused = !paused;
        input_next_frame( in );

        if( !paused ) {
            pos += accel;
            if( accel < accel_max ) {
                accel += accel_step;
            }
        } else {
            pos += accel;
            if( accel > 0 ) {
                accel -= accel_step;
            }
        }

        if( pos > (pagewidth << 16) ) {
            unsigned char *tmpp = page1;
            page1 = page2;
            page2 = tmpp;
            if( !fill_next_page( page2, pagewidth, pageheight, pagestride ) ) {
                fprintf( stderr, "viewcrawl: Can't open next crawl file, exiting.\n" );
                return 1;
            }
            pos -= (pagewidth << 16 );
        }

        /* Render image at the current position. */
        output->lock_output_buffer();
        for( i = 0; i < height; i++ ) {
            unsigned char *curscanline = output->get_output_buffer() + (stride * i);
            unsigned char *curpage1 = page1 + (i * pagestride);
            unsigned char *curpage2 = page2 + (i * pagestride);
            int pixpos = (pos >> 16);
            rgb32_subpix_blit_scanline( curscanline,
                                        curpage1 + (pixpos * 4),
                                        0, 0, 0,
                                        pos & 0xffff,
                                        pagewidth - pixpos );
            if( pos ) {
                unsigned char lastr = curpage1[ (pagewidth*4) - 2 ];
                unsigned char lastg = curpage1[ (pagewidth*4) - 3 ];
                unsigned char lastb = curpage1[ (pagewidth*4) - 4 ];
                rgb32_subpix_blit_scanline( curscanline + ((width - pixpos) * 4),
                                            curpage2, lastr, lastg, lastb,
                                            (pos & 0xffff), pixpos );
            }
        }
        output->unlock_output_buffer();

        /* Wait until blit time. */
        if( framecount % resync_frame == 0 ) {
            gettimeofday( &beforeblit, 0 );
            output->wait_for_retrace();
            gettimeofday( &afterblit, 0 );

            if( verbose ) {
                fprintf( stderr, "viewcrawl: Waited %d usec for the retrace.\n",
                         timediff( &afterblit, &beforeblit ) );
            }

            /* Reset our start time. */
            gettimeofday( &starttime, 0 );
            framecount = 0;
        }

        /* Sleep until it's time to blit the frame. */
        gettimeofday( &curtime, 0 );
        while( timediff( &curtime, &starttime )
               < ((framecount * usecperframe) + (usecperframe/4) - blittime - errortime) ) {

            if( rtctimer ) {
                rtctimer_next_tick( rtctimer );
            } else {
                /**
                 * Try to usleep for 1ms.  How well this will work
                 * unfortunately depends on your kernel.
                 */
                usleep( 1000 );
            }
            gettimeofday( &curtime, 0 );
        }

        /* Blit the frame. */
        gettimeofday( &beforeblit, 0 );
        output->show_frame();
        gettimeofday( &afterblit, 0 );
        blittime = timediff( &afterblit, &beforeblit );

        /* Report frame drops. */
        if( timediff( &afterblit, &lastblit ) > ((usecperframe*2)+(usecperframe/2)) ) {
            fprintf( stderr, "viewcrawl: Frame dropped?\n" );
        }

        if( verbose ) {
            fprintf( stderr, "viewcrawl: %d usec between blits, error %d usec.\n",
                     timediff( &afterblit, &lastblit ),
                     (framecount * usecperframe) + (usecperframe/4)
                             - timediff( &afterblit, &starttime ) );
        }

        lastblit = afterblit;
        framecount++;
    }

    output->shutdown();
    return 0;
}

